#! /usr/bin/env python
#
#  17bm_feedback.py -- IMCA-CAT 17-BM BPM feedback control program
#
#  Written by William Lavender.
#
#------------------------------------------------------------------------
#
#  Copyright 2004-2005 Illinois Institute of Technology
#
#  See the file "LICENSE" for information on usage and redistribution
#  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
#------------------------------------------------------------------------

class Servo:
	def __init__( self ):
		self.energy = None
		self.sp_coeff = None
		self.params = None
		self.kp = 0.0
		self.ki = 0.0
		self.kd = 0.0
		self.integral_limit = 1.0e38
		self.error_list = []
		self.max_error_integral_terms = 10
		self.last_error = None

	def add_measurement_to_error_list( self, measurement ):

		if (len(self.error_list) >= self.max_error_integral_terms):

			self.error_list = \
			    self.error_list[0:self.max_error_integral_terms - 1] 
		self.error_list = [ measurement ] + self.error_list

	def compute_error_integral( self ):
		scale = 1.0 / float(self.max_error_integral_terms)

		sum = 0.0

		error_list_length = len( self.error_list )

		for i in range(error_list_length):

#			print "error_list[%d] = %g" % ( i, self.error_list[i] )

			j = self.max_error_integral_terms - i

			sum = sum + \
			    float(j) * scale * float( self.error_list[i] )

		if ( sum > self.integral_limit ):
			sum = self.integral_limit

			print \
    "The servo loop attempted to exceed the positive integral limit", \
				self.integral_limit

		elif ( sum < -self.integral_limit ):
			sum = -self.integral_limit

			print \
    "The servo loop attempted to exceed the negative integral limit", \
				-self.integral_limit

		return sum

	def compute_correction( self, signal ):
		k0 = self.sp_coeff[0]
		k1 = self.sp_coeff[1]
		k2 = self.sp_coeff[2]
		k3 = self.sp_coeff[3]

		E = self.energy

		setpoint = k0 + ( k1 * E ) + ( k2 * E * E ) + ( k3 * E * E * E )

		error = signal - setpoint

		print "signal = %s, setpoint = %s, error = %s" \
			% ( signal, setpoint, error )

		print "kp = %s, ki = %s, kd = %s" \
			% ( self.kp, self.ki, self.kd )

		self.add_measurement_to_error_list( error )

		error_integral = self.compute_error_integral()

		print "error_integral =", error_integral

		try:
			if ( self.last_error == None ):
				error_difference = 0.0

				correction = - ( error * self.kp )
			else:
				error_difference = error - self.last_error

				correction = - ( error * self.kp \
				    + error_integral * self.ki \
				    + error_difference * self.kd )
		except:
			print "Error in servo calc.  Correction set to 0"

			correction = 0.0

		self.last_error = error

		print "error = %f, error_integral = %f, error_difference = %f" \
			% ( error, error_integral, error_difference )

		c0 = self.params[0]
		c1 = self.params[1]
		c2 = self.params[2]
		c3 = self.params[3]

		print "c0 = %g, c1 = %g, c2 = %g, c3 = %g" \
			% ( c0, c1 ,c2, c3)

		scaled_correction = c0 + c1 * correction \
				+ c2 * correction * correction \
				+ c3 * correction * correction * correction

		print "correction = %s, scaled_correction = %s" \
			% ( correction, scaled_correction )

		return scaled_correction

#------------------------------------------------------------------------

def print_interlock_status( in_progress_record, destination_record ):
	return

	in_progress = in_progress_record.read()
	destination = destination_record.read()

	print "cmd_in_progress = %g, cmd_destination = %g" \
		% ( in_progress, destination )

#------------------------------------------------------------------------

import sys, time

#  Add the Mp modules directory to the Python module load path.

sys.path[:0] = [ "/opt/mx/lib/mp" ]

import Mp

record_list = Mp.setup_database( "/opt/mx/etc/motor.dat" )

if ( len(sys.argv) != 1 ):
	print ""
	print "Usage:  %s" % ( sys.argv[0] )
	print ""
	sys.exit(0);

# Set tuning parameters.

print_debug = 1
save_values = 0

servo_loop_period = 0.1		;# In seconds

filename = "tempfile.dat"

# Sets how many times we average the signal in the feedback script.

num_qbpm_reads_record = record_list.get_record( "num_qbpm_reads" )

# Limits on the allowed sum signal.

bpm_sum_min_record = record_list.get_record( "bpm_sum_min" )

# Limits on the horizonal and vertical motion.

horiz_min_record = record_list.get_record( "bpm_x_min" )
horiz_max_record = record_list.get_record( "bpm_x_max" )

vert_min_record = record_list.get_record( "bpm_y_min" )
vert_max_record = record_list.get_record( "bpm_y_max" )

# Get MX beamline records used by the feedback.

x_fb_enable = record_list.get_record( "x_fb_enable" )
y_fb_enable = record_list.get_record( "y_fb_enable" )

vert_motor  = record_list.get_record( "mjack_pitch" )
horiz_motor = record_list.get_record( "chi" )

motors_tuple = ( vert_motor, horiz_motor )

posx = record_list.get_record( "bpm_ds_posx" )
posy = record_list.get_record( "bpm_ds_posy" )

energy_motor = record_list.get_record( "energy" )

x_setpoint_coeff = record_list.get_record( "bpm_x_setpoint" )
y_setpoint_coeff = record_list.get_record( "bpm_y_setpoint" )

x_params = record_list.get_record( "bpm_x_params" )
y_params = record_list.get_record( "bpm_y_params" )

x_kp = record_list.get_record( "bpm_x_kp" )
x_ki = record_list.get_record( "bpm_x_ki" )
x_kd = record_list.get_record( "bpm_x_kd" )

y_kp = record_list.get_record( "bpm_y_kp" )
y_ki = record_list.get_record( "bpm_y_ki" )
y_kd = record_list.get_record( "bpm_y_kd" )

curr1 = record_list.get_record( "bpm_ds_curr1" )
curr2 = record_list.get_record( "bpm_ds_curr2" )
curr3 = record_list.get_record( "bpm_ds_curr3" )
curr4 = record_list.get_record( "bpm_ds_curr4" )
bpm_sum = record_list.get_record( "bpm_ds_sum" )

cmd_in_progress = record_list.get_record( "cmd_in_progress" )
cmd_destination = record_list.get_record( "cmd_destination" )

shutter = record_list.get_record( "bshutter_closed" )

# Setup move deadbands for mjack_pitch and chi.

y_deadband_record = record_list.get_record( "bpm_y_deadband" )
x_deadband_record = record_list.get_record( "bpm_x_deadband" )

# Set up the servo loop parameters.

vert_servo = Servo()

vert_servo.kp = y_kp.read()
vert_servo.ki = y_ki.read()
vert_servo.kd = y_kd.read()

horiz_servo = Servo()

horiz_servo.kp = x_kp.read()
horiz_servo.ki = x_ki.read()
horiz_servo.kd = x_kd.read()

# Set integral limits for servo.

vert_servo.integral_limit = 1.0e5

horiz_servo.integral_limit = 1.0e5

# Open the file to save values to.

if save_values:
	try:
		file = open( filename, "w" )
	except:
		print "Cannot open file %s" % ( filename )
		sys.exit(1)

i = 1
pause_in_progress = 0

cmd_destination.write( 0 )
cmd_in_progress.write( 0 )

# print_interlock_status( cmd_in_progress, cmd_destination )

while (1):
	print "================================================"

	# Check to see if goniostat has requested that the feedback
	# loop pause.

	pause_requested_flag = cmd_in_progress.read()

#	print_interlock_status( cmd_in_progress, cmd_destination )

	if ( pause_requested_flag != 0 ):
		print "+++ Pause requested +++"
		pause_in_progress = cmd_destination.read()
		cmd_in_progress.write( 0 )
#		print_interlock_status( cmd_in_progress, cmd_destination )

	while ( pause_in_progress != 0 ):
		sys.stderr.write(".")

		pause_requested_flag = cmd_in_progress.read()
		pause_in_progress = cmd_destination.read()

#		print_interlock_status( cmd_in_progress, cmd_destination )

		if ( pause_requested_flag != 0 ):
			cmd_in_progress.write( 0 )

		if ( pause_in_progress == 0 ):
			print "+++ Pause complete +++"
			break

		time.sleep(0.01)

	# Check to see if the shutter is closed.  If it is closed, do not
	# do any feedback.

	shutter_closed = shutter.read()

	if ( shutter_closed != 0 ):
		print "+++ B shutter closed.  No feedback performed. +++"

		time.sleep(1.0)

		continue	# Go back to the top of the while loop.

	# Get current feedback limits.

	bpm_sum_min = bpm_sum_min_record.read()
	horiz_min = horiz_min_record.read()
	horiz_max = horiz_max_record.read()
	vert_min = vert_min_record.read()
	vert_max = vert_max_record.read()
	bpm_y_deadband = y_deadband_record.read()
	bpm_x_deadband = x_deadband_record.read()

	vert_motor.set_field( "raw_move_deadband", str(bpm_y_deadband) )
	horiz_motor.set_field( "raw_move_deadband", str(bpm_x_deadband) )

	if print_debug:
		print "x deadband = %s, y deadband = %s" \
			% ( horiz_motor.get_field( "raw_move_deadband" ), \
			vert_motor.get_field( "raw_move_deadband" ) )

	# Get the new setpoint.

	energy = energy_motor.get_position()

	vert_servo.energy = energy
	horiz_servo.energy = energy

	vert_servo.sp_coeff = y_setpoint_coeff.read()
	horiz_servo.sp_coeff = x_setpoint_coeff.read()

	vert_servo.params = y_params.read()
	horiz_servo.params = x_params.read()

	if print_debug:
		curr1_value = curr1.read()
		curr2_value = curr2.read()
		curr3_value = curr3.read()
		curr4_value = curr4.read()

	# Integrate BPM signals.

	num_qbpm_reads = num_qbpm_reads_record.read()

	vert_signal      = 0.0
	horiz_signal   = 0.0

	for n in range(num_qbpm_reads):
		posy_value = posy.read()
		posx_value = posx.read()

#		if print_debug:
#			print "n = %s, posx_value = %s, posy_value = %s" \
#				% ( n, posx_value, posy_value )

		vert_signal = vert_signal + posy_value
		horiz_signal = horiz_signal + posx_value
		time.sleep(0.01)

	vert_signal = vert_signal / float(num_qbpm_reads)
	horiz_signal = horiz_signal / float(num_qbpm_reads)

#	if print_debug:
#		print "1= %s, 3= %s, PY= %s,   4= %s, 2= %s, PX= %s" \
#		    % ( curr1_value, curr3_value, vert_signal, \
#			curr4_value, curr2_value, horiz_signal )

	# Get the current vertical and horizontal motor positions.

	vert_position  = vert_motor.get_position()
	horiz_position = horiz_motor.get_position()

	# Get the servo loop parameters

	vert_servo.kp = y_kp.read()
	vert_servo.ki = y_ki.read()
	vert_servo.kd = y_kd.read()

	horiz_servo.kp = x_kp.read()
	horiz_servo.ki = x_ki.read()
	horiz_servo.kd = x_kd.read()

	# Compute the servo correction.

	print "$$$ Computing Y Correction $$$"

	vert_correction = vert_servo.compute_correction( vert_signal )

	print "$$$ Computing X Correction $$$"

	horiz_correction = horiz_servo.compute_correction( horiz_signal )

	print "$$$ Done Computing Corrections $$$"

	new_vert_position = vert_position + vert_correction

	new_horiz_position = horiz_position + horiz_correction

#	if print_debug:
#		print "x = %s, y = %s" % ( horiz_signal, vert_signal )

	if save_values:
		file.write( "%d %f %f\n" % ( i, horiz_signal, vert_signal ) )
		file.flush()

	i = i+1

	new_vert_destination = new_vert_position

	if ( new_vert_destination > vert_max ):
		new_vert_destination = vert_max

	if ( new_vert_destination < vert_min ):
		new_vert_destination = vert_min

	new_horiz_destination = new_horiz_position

	if ( new_horiz_destination > horiz_max ):
		new_horiz_destination = horiz_max

	if ( new_horiz_destination < horiz_min ):
		new_horiz_destination = horiz_min

	# See if the signal is too small to do feedback with

	bpm_sum_value = bpm_sum.read()

	print "bpm_sum_value =", bpm_sum_value

	# If Y feedback is enabled, move the Y motor.

	y_enabled = y_fb_enable.read()

	if ( print_debug ) :
		print "1= %s, 3= %s, Pmjack_pitch= %s" \
		    % ( curr1_value, curr3_value, vert_signal )

		print "   mjack_pitch corr= %s, computed mjack_pitch = %s" \
			% ( vert_correction, new_vert_position )

		print "   old mjack_pitch = %s, new mjack_pitch= %s" \
			% ( vert_position, new_vert_destination )

	if ( y_enabled == 0 ):
		print "*** Y Feedback disabled ***"

	elif ( bpm_sum_value < bpm_sum_min ):
		print "*** BPM sum signal too small for Y feedback ***"

	else:
		print "Moving %s to %f" % \
			( vert_motor.name, new_vert_destination )

		vert_motor.move_absolute( new_vert_destination )

		print "***************************"

	# If X feedback is enabled, move the X motor.

	x_enabled = x_fb_enable.read()

	if ( print_debug ) :
		print "4= %s, 2= %s, Pchi= %s" \
		    % ( curr4_value, curr2_value, horiz_signal )

		print "   chi corr= %s, computed chi = %s" \
			% ( horiz_correction, new_horiz_position )

		print "   old chi = %s, new chi= %s" \
			% ( horiz_position, new_horiz_destination )

	if ( x_enabled == 0 ):
		print "*** X Feedback disabled ***"

	elif ( bpm_sum_value < bpm_sum_min ):
		print "*** BPM sum signal too small for X feedback ***"

	else:
		print "Moving %s to %f" % \
			( horiz_motor.name, new_horiz_destination )

		horiz_motor.move_absolute( new_horiz_destination )

		print "***************************"

	Mp.wait_for_multiple_motors_stop( motors_tuple )

	time.sleep(servo_loop_period)
